# code for using a sample, adjusting kappa function
# see UsingSamples.ipynb for runners
# (c) Stephen Mildenhall 2022

import aggregate as agg
from aggregate import build, Bounds, iman_conover, mu_sigma_from_mean_cv, random_corr_matrix
from collections import namedtuple
from functools import lru_cache
import logging
import numpy as np
import pandas as pd
from pandas.plotting import scatter_matrix
from scipy.sparse import coo_matrix
import scipy.stats as ss


logger = logging.getLogger(__name__)


def qdp(df):
    """
    quick describe with nice percentiles and cv
    """
    d = df.describe()
    # replace with non-sample sd
    d.loc['std'] = df.std(ddof=0)
    d.loc['cv'] = d.loc['std'] / d.loc['mean']
    return d


def make_test_sample(n, means, cvs, desired_correlation=None):
    """
    make a test DataFrame with sample lognormal marginals, given means and cvs

    """

    mus, sigmas = mu_sigma_from_mean_cv(means, cvs)
    c = len(mus)
    names = [chr(65+i) for i in range(c)]
    df = pd.DataFrame(
        {i: ss.lognorm(s,
                       scale=m*np.exp(-s*s/2)).rvs(n) for m, s, i in
         zip(means, sigmas, names)
         })
    if desired_correlation is not None:
        df = iman_conover(df, desired_correlation)
    return df


def portfolio_from_sample(n, means, cvs, rcm_p=1, positive=True, plot=True, log2=16, bs=1, **kwargs):
    """
    Create a sample and portfolio from the sample.
    Apply IC correlation to the sample.

    n = number of rows in sample
    rcm_p controls average magnitude of correlation
    positive = True for just positive correlation
    kwargs = passed through to update (bs=, log2=, remove_fuzz=)
    """
    c = len(means)
    if rcm_p is None:
        rcm = None
    elif isinstance(rcm_p, np.ndarray):
        rcm = rcm_p
    else:
        rcm = random_corr_matrix(c, rcm_p, positive)
    sample = make_test_sample(n, means, cvs, rcm)
    # display(sample.corr())
    if plot is True:
        scatter_matrix(sample, marker='.', s=5, alpha=1,
                       figsize=(10, 10), diagonal='kde')

    # add total etc.
    sample = sample.sort_index()
    sample = sample.reset_index(drop=False)

    # make the Portfolio object
    port = agg.Portfolio('test_from_df', sample)
    port.update(log2=log2, bs=bs, **kwargs)

    df_exa = add_exa_sample(port, sample)
    SampleResult = namedtuple(
        'SampleResult', ['sample', 'port', 'df_exa', 'rcm'])
    ans = SampleResult(sample, port, df_exa, rcm)
    return ans


def add_exa_sample(self, sample, S_calculation='forwards'):
    """
    Computes a version of density_df using sample to compute E[Xi | X].
    Then fill in the other ex.... variables using code from
    Portfolio.add_exa, stripped down to essentials.

    If no p_total is given then samples are assumed equally likely.
    total is added if not given (sum across rows)
    total is then aligned to the bucket size self.bs using (total/bs).round(0)*bs.
    The other loss columns are then scaled so they sum to the adjusted total

    Next, group by total, sum p_total and average the lines to create E[Xi|X]

    This sample is merged into a stripped down density_df. Then
    the other ex... columns are added. Excludes eta mu columns.

    Anticipated use: replace density_df with this, invalidate quantile
    function and then compute various allocation metrics.

    The index on sample is ignored.

    """

    # starter information
    # cut_eps = np.finfo(float).eps
    bs = self.bs

    # working copy
    sample_in = sample.copy()

    if 'total' not in sample:
        # p_total may be in sample
        cols = list(sample.columns)
        if 'p_total' in sample:
            cols.remove('p_total')
        sample_in['total'] = sample_in[cols].sum(axis=1)
    # index may be called total; that causes confusion; throw away input index
    sample_in = sample_in.reset_index(drop=True)

    # want to align the index to that of self.density_df; all multiples of self.bs
    # at the same time, want to scale all elements
    # temp0 gives the multiples of bs for the index; temp is the scaling for
    # all the other columns; temp0 will all be exact
    temp0 = (sample_in.total / bs).round(0) * bs
    temp = (temp0 / sample_in.total).to_numpy().reshape((len(sample_in), 1))
    # re-scale loss samples so they sum to total, need to extract p_total first
    if 'p_total' not in sample_in:
        # equally likely probs
        logger.info('Adding p_total to sample_in')
        # logger.info('Adding p_total to sample_in')
        p_total = 1.0 / len(sample_in)
    else:
        # use input probs
        p_total = sample_in['p_total']

    # re-scale
    sample_in = sample_in * temp
    # exact for total
    sample_in['total'] = temp0
    # and put probs back
    sample_in['p_total'] = p_total

    # Group by X values, aggregate probs and compute E[Xi  | X]
    exeqa_sample = sample_in.groupby(by='total').agg(
        **{f'exeqa_{i}': (i, np.mean) for i in self.line_names})
    # need to do this after rescaling to get correct (rounded) total values
    probs = sample_in.groupby(by='total').p_total.sum()
    # want all probs to be positive
    probs = np.maximum(0, probs.fillna(0.0))

    # working copy of self's density_df with relevant columns
    df = self.density_df.filter(
        regex=f'^(loss|(p|e)_({self.line_name_pipe})|(e|p)_total)$').copy()

    # want every value in sample_in.total to be in the index of df
    # this code verifies that has occurred
    # for t in sample_in.total:
    #     try:
    #         df.index.get_loc(t)
    #     except KeyError:
    #         print(f'key error for t={t}')
    #
    # or, if you prefer,
    #
    # test = df[['loss', 'p_total']].merge(sample_in, left_index=True, right_on='total', how='outer', indicator=True)
    # test.groupby('_merge')[['loss']].count()
    #
    # shows nothing right_only.

    # fix p_total and hence S and F
    # fill in these values (note, all this is to get an answer the same
    # shape as df, so it can be swapped in)
    df['p_total'] = probs
    df['p_total'] = df['p_total'].fillna(0.)

    # macro, F, S
    df['F'] = df.p_total.cumsum() # np.cumsum(df.p_total)

    if S_calculation == 'forwards':
        df['S'] = 1 - df.F
    else:
        # add_exa method; you'd think the fill value should be 0, which
        # will be the case when df.p_total sums to 1 (or more)
        df['S'] =  \
            df.p_total.shift(-1, fill_value=min(df.p_total.iloc[-1],
                                                max(0, 1. - (df.p_total.sum()))))[::-1].cumsum()[::-1]

    # this avoids irritations later on
    df.F = np.minimum(df.F, 1)
    df.S = np.minimum(df.S, 1)
    # where is S=0
    Seq0 = (df.S == 0)

    # TODO fix the means and other stats here? Only total changes!

    # invalidate quantile functions
    self._linear_quantile_function = None

    # E[X_i | X=a], E(xi eq a)
    # all in one go (outside loop)
    df = pd.merge(df,
                  exeqa_sample,
                  how='left',
                  left_on='loss',
                  right_on='total').fillna(0.0).set_index('loss', drop=False)
    # check exeqa sums to correct total. note this only happens ae, ie when
    # p_total > 0
    assert np.allclose(df.query('p_total > 0').loss,
                       df.query('p_total > 0')[[f'exeqa_{i}' for i in self.line_names]].sum(axis=1))
    # display(df.query('p_total > 0').head().T)

    # in another situation; here we are before exlea has been added
    # deal with the problem of conditioning on very small probabilities in the left tail
    # loss_max = df[['loss', 'exlea_total']].query('exlea_total > loss').loss.max()
    # if np.isnan(loss_max):
    #     loss_max = 0
    # else:
    #     # was mult * bs
    #     loss_max += bs
    #     logger.warning(f'Small probability fix: loss_max > 0, {loss_max}')
    # # try nan in place of 0             V
    # df.loc[0:loss_max, 'exlea_total'] = np.nan

    assert df.index.is_unique
    df['exeqa_total'] = df.loss

    # add additional variables via loop
    for col in self.line_names_ex:
        # ### Additional Variables
        # * exeqa_line = $E(X_i \mid X=a)$
        # * exlea_line = $E(X_i \mid X\le a)$
        # * e_line = $E(X_i)$
        # * exgta_line = $E(X_i \mid X \ge a)$
        # * exi_x_line = $E(X_i / X \mid X = a)$
        # * and similar for le and gt a
        # * exa_line = $E(X_i(a))$
        # * Price based on same constant ROE formula (later we will do $g$s)

        # need the stand alone LEV calc
        # E(min(Xi, a)
        # needs to be shifted down by one for the partial integrals....
        # stemp = 1 - df['p_' + col].cumsum()
        stemp = df['p_' + col].shift(-1, fill_value=min(df['p_' + col].iloc[-1],
                                                        max(0, 1. - (df['p_' + col].sum()))))[::-1].cumsum()[::-1]
        df['lev_' + col] = stemp.shift(1, fill_value=0).cumsum() * self.bs

        # E[X_i | X<= a] temp is used in le and gt calcs
        temp = np.cumsum(df['exeqa_' + col] * df.p_total)
        df['exlea_' + col] = temp / df.F

        # E[X_i | X>a]
        df['exgta_' + col] = (df['e_' + col] - temp) / df.S

        # E[X_i / X | X > a]  (note=a is trivial!)
        temp = df.loss.iloc[0]  # loss=0, should always be zero
        df.loss.iloc[0] = 1  # avoid divide by zero
        # unconditional E[X_i/X]
        df['exi_x_' + col] = np.sum(
            df['exeqa_' + col] * df.p_total / df.loss)
        temp_xi_x = np.cumsum(df['exeqa_' + col] * df.p_total / df.loss)
        df['exi_xlea_' + col] = temp_xi_x / df.F
        df.loc[0, 'exi_xlea_' + col] = 0  # selection, 0/0 problem
        # more generally F=0 error:                      V
        # df.loc[df.exlea_total == 0, 'exi_xlea_' + col] = 0
        # ?? not an issue for samples; don't have exlea_total anyway??
        # put value back
        df.loss.iloc[0] = temp

        fill_value = np.nan

        assert df.index.is_unique

        df['exi_xgta_' + col] = ((df[f'exeqa_{col}'] / df.loss *
                                  df.p_total).shift(-1, fill_value=fill_value)[
                                 ::-1].cumsum()) / df.S
        # need this NOT to be nan otherwise exa won't come out correctly
        df.loc[Seq0, 'exi_xgta_' + col] = 0.

        df['exi_xeqa_' + col] = df['exeqa_' + col] / df['loss']
        df.loc[0, 'exi_xeqa_' + col] = 0

        # need the loss cost with equal priority rule
        df[f'exa_{col}'] = (df.S * df['exi_xgta_' + col]).shift(1,
                                                                fill_value=0).cumsum() * self.bs

    # put in totals for the ratios... this is very handy in later use
    for metric in ['exi_xlea_', 'exi_xgta_', 'exi_xeqa_']:
        df[metric + 'sum'] = df.filter(regex=metric).sum(axis=1)

    df = df.set_index('loss', drop=False)
    df.index.name = None
    return df


def create_bounds(port, premium, a=0, n_tps=129, s=64):
    """
    THIS FUNCTION IS NOT NEEDED... use add_exa_sample bd = Bounds(df_exa)

    Summarize total and line's E[Xi|X] into data frame
    Set up bounds object

    line = line to which you want allocations...  [eventually, want to do them all...]

    Generic version to work with Portfolio object

    TODO: add asset limit?

    """

    raise ValueError('Routine not needed...why are you calling?')

    # :-6 takes out |total
    tot = port.density_df.filter(
        regex=f'loss|p_total|exeqa_({port.line_name_pipe})').copy()
    tot = tot.set_index('loss', drop=True)
    tot = tot.rename(columns={i: i[6:] for i in tot.filter(regex='exeqa_')})

    # p_total names the probabilities here
    tot = tot.groupby('total').agg(
        {**{'p_total': lambda x: np.sum(x)},
         **{l: np.mean for l in port.line_names_ex}}
    )

    # this is unnecessary for Agg generated...
    ptot = pd.DataFrame(index=port.density_df.index)
    ptot.index.name = 'total'
    ptot = ptot.merge(tot, how='left', left_index=True, right_index=True)
    ptot = ptot.fillna(0.)

    bd = Bounds(ptot)
    if a == 0:
        a = tot.index[-1]
    bd.tvar_cloud('total', premium=premium, a=a, n_tps=n_tps, s=s)
    p_star = bd.p_star('total', premium)

    return bd, ptot, p_star


def allocation_ranges(self, df_exa, a=np.inf):
    """
    price total and alloca te to other columns using to assets a
    across all calibrated biTVaRs

    TODO: fix when a < inf... need extra part of integral

    self = Bounds object

    df_exa = previously created index=loss X; p_total; E[Xi|X] X...

    :param a:
    :return:
    """

    # compute S
    # S = df_exa.query('p_total > 0').p_total.shift(-1, fill_value=0)[::-1].cumsum()[::-1]
    s = df_exa.S.to_numpy()

    hinges = coo_matrix(np.nan_to_num(np.minimum(1.0, s.reshape(
        1, len(s)) / (1.0 - self.tps.reshape(len(self.tps), 1))), nan=0.0))

    ml = coo_matrix((1 - self.weight_df.weight, (np.arange(len(self.weight_df)), self.idx[0])),
                    shape=(len(self.weight_df), len(self.tps)))
    mr = coo_matrix((self.weight_df.weight, (np.arange(len(self.weight_df)), self.idx[1])),
                    shape=(len(self.weight_df), len(self.tps)))
    m = ml + mr

    # print(f'm shape = {m.shape}, hinges shape = {hinges.shape}, types {type(m)}, {type(hinges)}')

    gS_cloud = (m @ hinges).toarray()

    # compute gp
    gps = -np.diff(gS_cloud, axis=1, prepend=1)

    # sum products for allocations
    bs = df_exa.index[1]
    bit = df_exa .filter(regex='exeqa_[A-Z]')
    # compute the allocations
    allocs = pd.DataFrame(
        gps @ (bit.to_numpy() * bs),
        columns=[i.replace('exeqa_', 'alloc_') for i in bit.columns],
        index=self.weight_df.index)

    return allocs


def plot_max_min(self, ax):
    ax.fill_between(self.cloud_df.index, self.cloud_df.min(
        1), self.cloud_df.max(1), facecolor='C7', alpha=.15)
    self.cloud_df.min(1).plot(ax=ax, label='_nolegend_', lw=0.5, ls='-', c='w')
    self.cloud_df.max(1).plot(ax=ax, label="_nolegend_", lw=0.5, ls='-', c='w')



# figures other than from PIR book

import numpy as np
# from numpy import roll
import scipy.stats as ss
from scipy.fft import rfft, irfft # , fftshift, ifftshift, fft, ifft
import matplotlib.pyplot as plt
from matplotlib import ticker
import matplotlib as mpl
from .. import build
from .. constants import FIG_H, FIG_W


def adjusting_layer_losses():
    """
    Figure to illustrate the process of adjusting layer losses.
    TODO: Add reference

    """
    f, ax = plt.subplots(1, 1, figsize=(FIG_H, FIG_W), constrained_layout=True)
    fz = ss.lognorm(.4)
    xs = np.linspace(0, fz.isf(1e-3), 1001, endpoint=False)
    F = fz.cdf(xs)
    ax.plot(F, xs)
    a1 = 1.5
    a0 = 1
    ax.axhline(a1, c='k', lw=.25)
    ax.axhline(a0, c='k', lw=.25)
    p0 = fz.cdf(a0)
    p1 = fz.cdf(a1)
    ax.plot([p0, p0], [0, a0], c='k', lw=.25)
    ax.plot([p1, p1], [0, a1], c='k', lw=.25)
    ax.set_xticks([0, p0, p1, 1])
    ax.set_xticklabels(['0', '$F(a_{i-1})$', '$F(a_{i})$', ''])
    ax.set_yticks([0, a0, a1, 3.5])
    ax.set_yticklabels(['0', '$a_{i-1}$', '$a_{i}$', ''])
    ax.text((p0+p1)/2, a0/2, '$f_i$', ha='center', va='center')
    ax.text((1+p1)/2, 1.25, '$e_i$', ha='center', va='center')
    xx = [.8]
    yy = [1.15]
    ll = ['$m_i$']
    ho = -0.2
    vo = 0.55
    ax.set(xlabel='Nonexceedance probability\n$F(x)$', ylabel='Outcome $x$',
           ylim=[0, 3.5], xlim=[-0.05, 1.05],
           title='Lee diagram\n$x$ vs $F(x)$')
    for x, y, l in zip(xx, yy, ll):
        ax.annotate(text=l,
                    xy=(x, y),
                    xytext=(x + ho, y + vo),
                    arrowprops={'arrowstyle': '->', 'linewidth': .5}
                    )
    return f


def savings_charge():
    """
    Figure to illustrate the insurance savings and expense(charge).

    """
    f, ax = plt.subplots(1, 1, figsize=(FIG_H, FIG_W), constrained_layout=True)
    fz = ss.lognorm(.4)
    xs = np.linspace(0, fz.isf(1e-3), 1001, endpoint=False)
    F = fz.cdf(xs)
    ax.plot(F, xs, lw=3)
    a1 = 1.25  # height of line
    ax.axhline(a1, c='C7', lw=1.5)
    ax.set_xticks([0, .25, .5, .75, 1])
    ax.set_yticks([0, a1, 3.5])
    ax.set_yticklabels(['0', '$r$', ''])
    ax.set(xlabel='Nonexceedance probability', ylabel='Scaled outcome',
           ylim=[0, 3.5], xlim=[0, 1],
           title='Insurance savings and expense')

    xx = [.9, .25]
    yy = [1.05 * a1, .8* a1]
    ll = ['Insurance\ncharge, $\\phi(r)$',
         'Insurance\nsavings, $\\psi(r)$']
    hos = [-0.1, 0.25]
    vos = [1.2, .65]
    ax.text(.5, a1 * .4, '$E[X\\wedge l]$', ha='center', va='center')
    for x, y, l, ho, vo in zip(xx, yy, ll, hos, vos):
        ax.annotate(text=l,
                    xy=(x, y),
                    xytext=(x + ho, y + vo),
                    ha='right',
                    va='bottom',
                    arrowprops={'arrowstyle': '->', 'linewidth': .5}
                    )
    return f


def mixing_convergence(freq_cv, sev_cv, bs=1/64):
    """
    Illustrate convergence of mixed distributions to the mixing distribution.

    """

    # make the two dataframes of distributions
    a = build('agg M '
              '1 claims '
              f'sev gamma 1 cv {sev_cv} '
              f'mixed gamma {freq_cv} ', log2=16, bs=bs, approximation='exact')
    dfnb = a.density_df[['p']].rename(columns={'p': 1})
    assert np.abs(a.est_m / a.agg_m - 1) < 1e-3
    # print("1", a.agg_m, a.est_m, a.est_m / a.agg_m - 1)

    for freq in [2, 5, 10, 20, 50, 100, 200]:
        a = build('agg M '
                  f'{freq} claims '
                  f'sev gamma 1 cv {sev_cv} '
                  f'mixed gamma {freq_cv} ', log2=16, bs=bs, approximation='exact')
        dfnb[freq] = a.density_df[['p']]
        assert np.abs(a.est_m / a.agg_m - 1) < 1e-3
        # print(freq, a.agg_m, a.est_m, a.est_m / a.agg_m - 1)

    a = build('agg M '
              '1 claims '
              f'sev gamma 1 cv {sev_cv} '
              'poisson ', log2=16, bs=bs, approximation='exact')
    dfp = a.density_df[['p']].rename(columns={'p': 1})
    assert np.abs(a.est_m / a.agg_m - 1) < 1e-3
    # print("1", a.agg_m, a.est_m, a.est_m / a.agg_m - 1)

    for freq in [2, 5, 10, 20, 50, 100, 200]:
        a = build('agg M '
                  f'{freq} claims '
                  f'sev gamma 1 cv {sev_cv} '
                  'poisson ', log2=16, bs=bs, approximation='exact')
        dfp[freq] = a.density_df[['p']]
        assert np.abs(a.est_m / a.agg_m - 1) < 1e-3
        # print(freq, a.agg_m, a.est_m, a.est_m / a.agg_m - 1)

    # plotting
    fig, axs = plt.subplots(2, 2, figsize=(
        2 * FIG_W, 2 * FIG_H), constrained_layout=True)
    axi = iter(axs.flat)

    for lbl, df in zip(['Poisson distribution', 'Mixed frequency distribution'], [dfp, dfnb]):
        ax0 = next(axi)
        ax1 = next(axi)
        for c in df:
            ax0.plot(df.index/c, c*df[c],       lw=1, label=str(c))
            ax1.plot(df.index/c, df[c].cumsum(), lw=1, label=str(c))

        if ax0 is axs.flat[0]:
            ax0.legend()
        ax0.set(ylim=[0, 1e-1], xlim=[-0.25, 5])
        ax1.set(ylim=[-0.05, 1.05], xlim=[-0.25, 5])

        ax0.set(ylabel='Density', xlabel='Normalized loss', title=lbl)
        ax1.set(ylabel='Distribution', xlabel='Normalized loss', title=lbl)

    # add convergence to mixing
    alpha = freq_cv**-2
    fz = ss.gamma(alpha, scale=1/alpha)
    ps = np.linspace(0, 5, 501)
    ax = axs.flat[-1]
    ax.plot(ps, fz.cdf(ps), lw=2, alpha=.5, c='k', label='Mixing')
    ax.legend(loc='lower right')


def power_variance_family():
    """
    Graph to illustrate the power variance exponential family distributions.

    Reference: Jørgensen, Bent. 1997. The theory of dispersion models. CRC Press.
    """
    alpha = np.linspace(-2, 2, 101)
    p = (alpha-2) / (alpha-1)
    alphabar = -(alpha+1)
    f, ax = plt.subplots(figsize=(FIG_W * 2, FIG_W * 2))
    ax.plot(alphabar, p, lw=3)
    ax.set(ylim=[-5,10])
    # ax.grid(lw=.25, c='b', alpha=.5)
    ax.axhline(1, c='k', lw=1)
    ax.axhline(0, c='k', lw=1)
    ax.axvline(-2, c='k', lw=.5)
    ax.axvline(-3/2, c='k', lw=0.5, ls='--')
    ax.axhline(3, c='k', lw=0.5, ls='--')
    ax.axhline(2, c='r', lw=1)
    ax.axvline(-1, c='r', lw=1)
    ax.set(xlabel='$\\bar\\alpha=-(\\alpha+1)$, base jump density is $x^{\\bar\\alpha}$',
           ylabel='Variance power function $p$, $V(\\mu)=\\phi\\mu^p$')
    ax.yaxis.set_major_locator(ticker.FixedLocator(np.arange(-4,10)))

    def ql(x, y, t, dot=True, rhs=None):
        ax.text(x + .05, y+0.2, t)
        if dot:
            ax.plot(x,y, 'rd', ms=5)
        else:
            if rhs is None:
                rhs = x + 1
            ax.plot([x, rhs], [y, y], lw=3)

    ql(-3, 0, 'Normal')
    ql(-3, 6, 'Stable', False)
    ql(-2, 9, 'Cauchy')
    ql(-2, -2, 'Pos Stable', False)
    ql(-1.5, 3, 'Stab 3/2\nIG')
    ql(-1, 2, 'Gamma')
    ql(1, 1, 'Poisson')
    ql(-1, -2, 'Tweedie', False, 1)
    ax.set(title='Power Variance Exponential Family Distributions');




# make the tame case study 

from aggregate.extensions import case_studies as cs
from aggregate import build

if __name__ == '__main__':

    # calibrate reinsurance
    recalc = build('agg B 1 claim sev gamma 50 cv 0.15 fixed',
                   log2=16, bs=1/64)
    a, d = recalc.q(0.8), recalc.q(0.99)
    y = d - a

    # create exhibits
    tame = cs.CaseStudy()
    tame.factory(case_id='tame',
                 case_name='Tame',
                 case_description='Tame Case in the new syntax.',
                 a_distribution='agg A 1 claim sev gamma  50 cv 0.10 fixed',
                 b_distribution_gross='agg B 1 claim sev gamma  50 cv 0.15 fixed',
                 b_distribution_net=f'agg B 1 claim sev gamma  50 cv 0.15 fixed aggregate net of {y} xs {a}',
                 reg_p=0.9999,
                 roe=0.10,
                 d2tc=0.3,
                 f_discrete=False,
                 s_values=[.005, 0.01, 0.03],
                 gs_values=[0.029126,   0.047619,   0.074074],
                 bs=1/64,
                 log2=16,
                 padding=1)
    tame.full_monty()
    tame.to_json()
    tame.browse_exhibits()



# risk progression functions: decomposition of the natural allocation

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import namedtuple
from aggregate import make_mosaic_figure


def ff(n):
    s = f'{{x:.{n}f}}'
    pd.set_option('float_format', lambda x: s.format(x=x))


def make_distribution(df, prob_column='p_total'):
    """
    Make df, with columns [something, prob_column], into a distribution: groupby
    something, sum probs. Return the distribution and the sum of probs.
    """
    c = list(df.columns)
    c.remove(prob_column)
    out = df.set_index(c[0]).groupby(c[0]).sum()[prob_column]
    sp = out.sum()
    # normalize?
    out = out / sp
    return out, sp


def make_projection_distributions(self):
    """
    self = Portfolio
    make distributions from exeqa projections.

    """
    projections = {}
    sum_probs = {}
    for unit in self.unit_names:
        proj, sum_probs[unit] = \
            make_distribution(self.density_df[[f'exeqa_{unit}', 'p_total']])
        proj = proj.to_frame()
        proj['F'] = proj.p_total.cumsum()
        proj['S'] = 1 - proj.F
        projections[unit] = proj

    return projections, sum_probs


def plot_comparison(self, projections, axs, smooth):
    """
    Plot to compare unit, projection of unit, and total in self.
    self is a Portfolio

    :param self: Portfolio
    :param projections: dict of projections
    :param axs: list of axes
    :param smooth: smoothing factor for densities
    """
    lw = 2
    for unit, proj, axd, ax, axr, lc in \
        zip(self.unit_names, projections.values(), axs.flat[::3],
            axs.flat[1::3], axs.flat[2::3], ['C1', 'C2']):

        # figure the mean
        mn = np.sum(proj.p_total * proj.index)
        # print(
        #     f'{unit}, agg est mean {self[unit].est_m}, sum prod mean {mn}')

        # normalized densities: exeqa
        p = proj.p_total.copy()
        # rebucket to smooth out
        quant = self.bs * smooth
        p.index = quant * np.round(p.index / quant)
        p = p.groupby(level=0).sum()
        p.iloc[:-1] = p.iloc[:-1] / np.diff(p.index) * mn
        p.index = p.index / mn
        p.plot(ax=axd, c=lc, ls='--', label='Projection of ' + unit)

        # original density
        p = self[unit].density_df.p_total.copy()
        p.index = p.index / self[unit].est_m
        p = p * self[unit].est_m / self.bs
        p.plot(ax=axd, c=lc, label=unit)

        # total density
        p = self.density_df.p_total.copy()
        p.index = p.index / self.est_m
        p = p * self.est_m / self.bs
        p.plot(ax=axd, c='C0', label='total')
        # remember: normalized so 0 to 5 is universal
        axd.set(xlim=[0, 5],
                xlabel='loss', ylabel='density')
        yl = axd.get_ylim()
        if yl[1] > 100:
            axd.set(yscale='log', ylabel='log density')
        axd.legend(loc='upper right').set(title='NORMALIZED losses')

        # plot normalized distributions on linear and return period scale
        ax.plot(self.density_df[f'p_{unit}'].cumsum(),
                self.density_df.loss / self[unit].est_m, c=lc, lw=lw/2, label=unit)
        ax.plot(proj.F, np.array(proj.index) / mn,
                c=lc, ls='--', lw=lw, label='Projection')
        ax.plot(self.density_df['F'], self.density_df.loss /
                self.est_m, c='C0', lw=lw, label='total')
        ax.set(ylim=[0, 5], xlabel='probability', ylabel='normalized loss')
        ax.axhline(1, lw=.5, ls='--', c='C7')
        ax.legend(loc='upper left')

        axr.plot(1 / (1 - self.density_df[f'p_{unit}'].cumsum()),
                 self.density_df.loss / self[unit].est_m, c=lc, lw=lw/2, label=unit)
        proj = proj.query('F > 1e-11 and S > 1e-11')
        axr.plot(1 / proj.S, np.array(proj.index) / mn,
                 c=lc, ls='--', lw=lw, label='Projection')
        axr.plot(1 / self.density_df['S'], self.density_df.loss /
                 self.est_m, c='C0', lw=lw, label='total')
        axr.set(xlim=[1, 1e4], ylim=1e-1, xscale='log', yscale='log',
                xlabel='log return period', ylabel='log normalized loss')
        axr.axhline(1, lw=.5, ls='--', c='C7')
        axr.legend(loc='lower right')


def make_up_down(ser):
    """
    Split ser, regarded as a function on its index, into its increasing and decreasing
    parts. Also return the reconstruction.

    [was difference_increasing]

    Example::

        df = pd.DataFrame({'x': np.linspace(0, 20, 2001)})
        df['y'] = 2 * np.sin(df.x) + 1.8 * np.cos(df.x * 3 + 2)
        df = df.set_index('x')
        u,d,c = make_up_down(df.y)
        ax = u.plot()
        d.plot(ax=ax)
        c.plot(ax=ax)
        df.y.plot(ax=ax, lw=.5, c='w')
        ax.legend()

    :param ser: pandas Series
    """
    dy = ser.diff()
    dy.iloc[0] = ser.iloc[0]
    u = np.maximum(0, dy).cumsum()
    d = -np.minimum(0, dy).cumsum()
    c = u - d
    u.name = 'up'
    d.name = 'down'
    c.name = 'recreated'
    return u, d, c


def up_down_distributions(self):
    """
    Write the projections as differences of increasing functions and
    figure the distributions of those two functions.
    self is a Portfolio
    """
    up_functions = {}
    down_functions = {}
    up_distributions = {}
    down_distributions = {}
    for unit in self.unit_names:
        u, d, c = make_up_down(self.density_df[f'exeqa_{unit}'])
        up_functions[unit] = u
        down_functions[unit] = d

        u = u.to_frame()
        u['p_total'] = self.density_df.p_total
        du, _ = make_distribution(u)
        up_distributions[unit] = du

        d = d.to_frame()
        d['p_total'] = self.density_df.p_total
        dd, _ = make_distribution(d)
        down_distributions[unit] = dd

    UDD = namedtuple('UDD', ['up_functions', 'down_functions',
                             'up_distributions', 'down_distributions'])
    ans = UDD(up_functions, down_functions,
              up_distributions, down_distributions)
    return ans


def plot_up_down(self, udd, axs):
    """
    Plot the decomposition of the projections into up and down parts
    self - portfolio
    udd = UDD named tuple (above)
    """

    for unit, ax in zip(self.unit_names, axs.flat):
        ax = self.density_df[f'exeqa_{unit}'].plot(ax=ax, lw=4, c='C7')
        udd.up_functions[unit].plot(ax=ax)
        udd.down_functions[unit].plot(ax=ax)
        (udd.up_functions[unit] - udd.down_functions[unit]
         ).plot(ax=ax, lw=1.5, ls=':', c='C2', label='recreated')
        ax.legend()
        ax.set(xlabel='loss', ylabel='up or down function')

    # plot ud distributions
    ax = axs.flat[-1]
    for (k, v), c in zip(udd.up_distributions.items(), ['C0', 'C1']):
        v.cumsum().plot(c=c, ax=ax, label=f'Up {k}')
    for (k, v), c in zip(udd.down_distributions.items(), ['C0', 'C1']):
        v.cumsum().plot(c=c, ls=':', ax=ax, label=f'Down {k}')
    ax.legend(loc='lower right')
    ax.set(xlabel='loss', ylabel='cumulative probability')


def price_work(dn, series, names_ex):
    """
    price using dn across a number of series (already distributions) with given
    assets (that tie back to probs... caller OR pass in Port? )

    NO ASSET VERSION
    uses dn.price not price2

    """

    bit0 = pd.concat([
        pd.DataFrame(dn.price(ser, np.inf, kind='both'),
                     index=['bid', 'el', 'ask'], columns=[1]).T
        for ser in series],
        keys=names_ex, names=['unit', 'view'])
    bit = bit0.stack().unstack('view')
    # add sum of parts
    bit.loc[('sum', 'bid'), :] = 0.0
    bit.loc[('sum', 'el'), :] = 0.0
    bit.loc[('sum', 'ask'), :] = 0.0
    # TODO Kludge
    bit.loc['sum'] = bit.loc[names_ex[0]].values + bit.loc[names_ex[1]].values

    return bit


def price_compare(self, dn, projection_dists, ud_dists):
    """
    Build out pricing comparison waterfall

    :param self: Portfolio
    :param dn: Distortion
    """

    # linear natural allocation pricing
    # KLUDGE
    lna_pricea = self.price(self.q(1), dn, view='ask')
    # display(lna_pricea.df)
    lna_priceb = self.price(self.q(1), dn, view='bid')
    na_price = lna_pricea.df[['L', 'P']]
    na_price.columns = ['el', 'ask']
    # if this is nan it gets dropped from stack?!
    na_price['bid'] = lna_priceb.df['P']
    na_price = na_price.stack(dropna=False).to_frame()
    # beware : order matters here
    na_price.loc[('sum', 'el'), :] = 0.0
    na_price.loc[('sum', 'ask'), :] = 0.0
    na_price.loc[('sum', 'bid'), :] = 0.0
    na_price.loc['sum'] = na_price.loc[self.unit_names[0]
                                       ].values + na_price.loc[self.unit_names[1]].values
    # call pricing function FOUR times; original dist; proj; u and d of proj

    # original dists = stand alone pricing
    series = [self.density_df[f'p_{unit}'] for unit in self.unit_names_ex]
    sa = price_work(dn, series, self.unit_names_ex)

    # projected sa
    series = [v['p_total']
              for v in projection_dists.values()] + [self.density_df.p_total]
    proj_sa = price_work(dn, series, self.unit_names_ex)

    # up and down sa
    series = [v for v in ud_dists.up_distributions.values()] + \
        [self.density_df.p_total]
    udu = price_work(dn, series, self.unit_names_ex)
    udu.loc['total'] = np.nan
    # up and down sa
    series = [v for v in ud_dists.down_distributions.values()] + \
        [self.density_df.p_total]
    udd = price_work(dn, series, self.unit_names_ex)
    udd.loc['total'] = np.nan

    compare = pd.concat([
        na_price, sa, proj_sa, udu, udd
    ],
        axis=1,
        keys=['lna', 'sa', 'proj_sa', 'up', 'down']).droplevel(1, 1)

    # put in ask, el, bid order
    compare = compare.iloc[np.hstack(
        [np.array([0, 2, 1]) + 3*i for i in range(4)])]
    return compare


def full_monty(self, dn, truncate=True, smooth=16):
    """
    One-stop shop for a Portfolio self
    Unlimited assets
    Prints all on one giant figure
    """

    # figure for all plots
    fig, axs = plt.subplots(4, 3, figsize=(
        3 * 3.5, 4 * 2.45), constrained_layout=True)

    # in the known bounded case we can truncate
    regex = ''.join([i[0] for i in self.line_names_ex])
    if truncate:
        self.density_df = self.density_df.loc[:self.density_df.F.idxmax()]
        self._linear_quantile_function = None

    # density and exa plots
    axd = {'A': axs[0, 0], 'B': axs[0, 1], 'C': axs[0, 2]}
    self.plot(axd=axd)
    self.density_df.filter(regex=f'exeqa_[{regex}]').plot(ax=axd['C'])
    axd['C'].set(xlabel='loss', ylabel='Conditional expectation')

    # projection distributions
    projection_dists, sum_probs = make_projection_distributions(self)
    if not np.allclose(list(sum_probs.values()), 1):
        print(sum_probs)

    # impact of projections on distributions
    axs1 = axs[1:3, :]
    plot_comparison(self, projection_dists, axs1, smooth)

    # up and down decomp
    ud_dists = up_down_distributions(self)

    # plot UD
    axs1 = axs[3, :]
    plot_up_down(self, ud_dists, axs1)

    compare = price_compare(self, dn, projection_dists, ud_dists)
    compare['umd'] = compare['up'] - compare['down']

    RiskProgression = namedtuple('RiskProgression', ['compare_df', 'projection_dists', 'ud_dists'])
    ans = RiskProgression(compare, projection_dists, ud_dists)
    return ans



# code for running test cases, producing HTML, etc.

from .. import pprint_ex

# from ..aggregate.utilities  import iman_conover, mu_sigma_from_mean_cv
# # from aggregate.utils import rearrangement_algorithm_max_VaR
# from .. aggregate.utilities import random_corr_matrix
import logging
import matplotlib.pyplot as plt
from pathlib import Path
import re


logger = logging.getLogger(__name__)


class TestSuite(object):

    p = None
    build = None
    tests = ''

    @classmethod
    def __init__(cls, build=None, out_dir_name=''):
        """
        Run test suite fn. Create specified objects. Save graphics and info to HTML. Wrap
        HTML with template.

        TODO: convert wrapping to Jinja!

        To run whole test_suite
        ::

            python -m aggregate.extensions.test_suite

        """

        if build is None:
            from .. import build

        cls.build = build

        if out_dir_name != '':
            cls.p = Path(out_dir_name)
            if cls.p.exists() is False:
                raise FileExistsError(f'Directory {out_dir_name} does not exist.')
        else:
            cls.p = cls.build.site_dir.parent / 'generated'
            cls.p.mkdir(exist_ok=True)
            (cls.p / "img").mkdir(exist_ok=True)

        logger.info(f'Output directory {cls.p.resolve()}')

        # extract from comments; this is just FYI
        fn = 'test_suite.agg'
        suite = build.default_dir / fn
        txt = suite.read_text(encoding='utf-8')
        tests = [i for i in txt.split('\n') if re.match(r'# [A-Z]\.', i)]
        cls.tests = [i.replace("# ", "").split('. ') for i in tests]

    @classmethod
    def run(cls, regex, title, fig_prefix, fig_format='svg', fig_size=(8,2.4), **kwargs):
        """

        :param regex: regex of tests to run, e.g., 'agg [ABC]\. '
        :param title: title for blob
        :param fig_prefix: file name prefix for saved immage files (convenience)
        :param fig_format:  html or markdown (md); html uses svg output, markdown uses pdf
        :param fig_size:
        :param kwargs: passed to savefig
        """
        logger.warning(f'figure prefix = {fig_prefix}')

        ans = []
        for n in cls.build.qshow(regex).index:
            a = cls.build(n)
            ans.append(a.html_info_blob().replace('h3>', 'h2>'))
            ans.append(pprint_ex(a.program, 50, True, True))
            ans.append(cls.style_df(a.describe).to_html())
            ans.append('<br>')
            fn = cls.p / f'img/{fig_prefix}_tmp_{hash(a):0x}.{fig_format}'
            a.plot(figsize=fig_size)
            a.figure.savefig(fn, **kwargs)
            ans.append(f'<img src="{fn.resolve()}" />')
            plt.close(a.figure)
            logger.warning(f'Created {n}, mean {a.agg_m:.2f}')

        blob = '\n'.join(ans)
        fn = cls.p / f'{fig_prefix}.html'
        fn.write_text(blob, encoding='utf-8')

        fn2 = cls.p / f'{fn.stem}_wrapped.html'
        fn3 = cls.build.template_dir / 'test_suite_template.html'
        # TODO JINJA!
        template = fn3.read_text()
        template = template.replace('HEADING GOES HERE', title).replace(
            'CONTENTHERE', blob)
        fn2.write_text(template, encoding='utf-8')

    @staticmethod
    def style_df(df):
        """
        Style a df similar to pricinginsurancerisk.com styles.

        graph background color is B4C3DC and figure (paler) background is F1F8F#

        Dropped row lines; bold level0, caption

        :param df:
        :return: styled dataframe

        """

        cell_hover = {
            'selector': 'td:hover',
            'props': [('background-color', '#ffffb3')]
        }
        index_names = {
            'selector': '.index_name',
            'props': 'font-style: italic; color: black; background-color: white; '
                     'font-weight:bold; border: 0px solid #a4b3dc; text-transform: capitalize; '
                     'text-align:left;'
        }
        headers = {
            'selector': 'th:not(.index_name)',
            'props': 'background-color: #DDDDDD; color: black;  border: 1px solid #ffffff;'
        }
        center_heading = {
            'selector': 'th.col_heading',
            'props': 'text-align: center;'
        }
        left_index = {
            'selector': '.row_heading',
            'props': 'text-align: left;'
        }
        td = {
            'selector': 'td',
            'props': f'text-align: right; '
        }
        nrow = {
            'selector': 'tr:nth-child(even)',
            'props': 'background-color: #F5F5F5;'
        }
        all_styles = [cell_hover, index_names, headers, center_heading, nrow, left_index, td]

        fc = lambda x: f'{x:,.3f}' if isinstance(x, (float, int)) else x
        f3 = lambda x: f'{x:.3f}' if isinstance(x, (float, int)) else x
        f5g = lambda x: f'{x:.5g}' if isinstance(x, (float, int)) else x
        # guess sensible defaults
        fmts = {'E[X]': fc,
                'Est E[X]': fc,
                'Err E[X]': f5g,
                'CV(X)': f3,
                'Est CV(X)': f3,
                'Err CV(X)': f5g,
                'Skew(X)': f3,
                'Est Skew(X)': f3}
        return df.style.set_table_styles(all_styles).format(fmts)


def run_test_suite():
    t = TestSuite()
    # show progress
    t.build.logger_level(30)
    print(t.tests)
    # run all the aggs
    # TODO FIX for Portfolios
    # t.run(regex=r'^C\.', title='C only', fig_prefix="auto", fig_format='png', dpi=300)
    t.run(regex=r'^[A-KNO]\.', title='Full Test Suite', fig_prefix="auto", fig_format='png', dpi=300)


if __name__ == '__main__':
    run_test_suite()



# coding: utf-8 -*-

from . parser import UnderwritingLexer, UnderwritingParser, grammar
from . utilities import get_fmts, pprint, pprint_ex, ft, \
    ift, sln_fit, sgamma_fit, estimate_agg_percentile, \
    axiter_factory, AxisManager, lognorm_lev, html_title, \
    sensible_jump, suptitle_and_tight, \
    MomentAggregator, MomentWrangler, xsden_to_meancv, xsden_to_meancvskew, \
    frequency_examples, Answer, log_test, subsets, \
    round_bucket, \
    make_ceder_netter, make_mosaic_figure, nice_multiple, \
    style_df, logger_level, friendly, \
    FigureManager, tweedie_convert, tweedie_density,\
    iman_conover, rearrangement_algorithm_max_VaR, \
    mu_sigma_from_mean_cv, \
    make_corr_matrix, random_corr_matrix, \
    LoggerManager, knobble_fonts, approximate_work, \
    partial_e, partial_e_numeric, moms_analytic, qd, \
    sEngFormatter, mv, picks_work, GCN, lognorm_approx, \
    integral_by_doubling
from . spectral import Distortion
from . distributions import Frequency, Severity, Aggregate
from . portfolio import Portfolio, make_awkward
from . underwriter import Underwriter, build, debug_build
from . bounds import Bounds, plot_max_min, plot_lee
from . constants import *

import sys

# knobble warnings
# https://docs.python.org/3/library/warnings.html#temporarily-suppressing-warnings
if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")


__docformat__ = 'restructuredtext'
__project__ = 'aggregate'
__author__ = "Stephen J. Mildenhall"
__copyright__ = "2018-2022, Convex Risk LLC"
__license__ = "BSD 3-Clause New License"
__email__ = "steve@convexrisk.com"
__status__ = "alpha"
# only need to change here, feeds conf.py (docs) and setup.py (build)
__version__ = "0.11.1"

# set up
from pathlib import Path
base_dir = Path.home() / 'aggregate'
base_dir.mkdir(exist_ok=True)

for p in ['cases', 'parser', 'temp', 'generated']:
    (base_dir / p).mkdir(exist_ok=True)

# print('All directories exist')

del p, base_dir


# imports

# as a default turn off all logging
logger_level(30)
knobble_fonts()

# module level doc-string
__doc__ = """
:mod:`aggregate` solves insurance, risk management, and actuarial problems using realistic models that reflect underlying frequency and severity. It makes working with an aggregate (compound) probability distribution as easy as the lognormal, delivering the speed and accuracy of parametric distributions to situations that usually require simulation. :mod:`aggregate` includes an expressive language called DecL to describe aggregate distributions and is implemented in Python under an open source BSD-license.
"""
